--MATH.E v1.3 by Andy Montgomery
--Mathematical routines that work on atoms or seqences. Also a few constants.

--Functions:
--arccos() - inverse cosine
--arcsin() - inverse sine
--sign() - returns -1 for negative numbers, 0 for 0, and 1 for positives.
--abs() - absolute value
--cosh() - hyperbolic cosine
--sinh() - hyperbolic sine
--tanh() - hyperbolic tangent
--arccosh() - inverse cosh
--arcsinh() - inverse sinh
--arctanh() - inverse tanh
--int() - greatest integer
--ln() - natural logarithm (base e)
--logbase(b,a) - logarithm base b of a
--log10() - common logarithm (base 10)
--sortA() - sorts a sequence in ascending order
--sortD() - sorts a sequence in descending order
--max() - highest value of a sequence (also accepts atoms, but why?)
--min() - lowest value of a sequence (ditto)
--mean() - average of a sequence, averages nested sequences first for weighting
--median() - median of a sequence, middle value of sorted sequence
--variance() - variance of a sequence
--std_dev() - standard deviation of a sequence
--oned() - converts atoms or multidimentional sequences to 1D sequences
--fib() - returns the nth Fibonanci number
--factorial() - factorial
--nPr() - number of permutations
--nCr() - number of combinations

--Check out the formula I used in fib().  I was flabbergasted when my math
--teacher showed it to us.  It's not recursive!

--Constants:
--c_e is e, c_pi is pi, c_gr is the golden ratio, (1+sqrt(5))/2, c_grneg is
--a number related to the golden ratio that is used in the fib() formula,
--nan is the "not a number" the Euphoria documentation talks about, inf
--is the "infinity" it mentions.

--Future improvements: more sequence manipulation routines, matrix operations,
--prime factorization, rounding, greatest common factor, lowest common
--multiple, better modular division than remainder(), numerical (not symbolic)
--calculus functions, complex numbers, vectors, choice of exact answer or
--decimal approximation, combinations and permutations, polynomial root
--finder, decimal to fraction converter, anything else I think of.

without warning

global constant c_e=2.718281828459,
                c_pi=3.1415926535898,
                c_gr=1.6180339887499,
                c_grneg=-0.6180339887499,
                nan=-(2*power(10,308)/(3*power(10,308))),
                inf=2*power(10,308)

global function arccos(object x)
object result
if sequence(x) then
    result=repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = arccos(x[count])
    end for
else
    if (x>=-1 and x<=1) then
        result = c_pi/2 - 2*arctan(x/(sqrt(1.0 - x*x) + 1.0))
    else
        result = nan
    end if
end if
return result
end function

global function arcsin(object x)
object result
if sequence(x) then
    result=repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = arcsin(x[count])
    end for
else
    if (x>=-1 and x<=1) then
        result = 2*arctan(x/(sqrt(1.0 - x*x) + 1.0))
    else
        result = nan
    end if
end if
return result
end function

global function sign(object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = sign(x[count])
    end for
else
    if x<0 then
        result=-1
    elsif x>0 then
        result=1
    elsif x=0 then
        result=0
    else
        result=nan
    end if
end if
return result
end function

global function abs(object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = abs(x[count])
    end for
else
    if x<0 then
        result = -x
    else
        result = x
    end if
end if
return result
end function

global function cosh(object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = cosh(x[count])
    end for
else
    result = power(c_e,x)/2 + power(c_e,-x)/2
end if
return result
end function

global function sinh(object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = sinh(x[count])
    end for
else
    result = power(c_e,x)/2 - power(c_e,-x)/2
end if
return result
end function

global function tanh(object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = tanh(x[count])
    end for
else
    result = (power(c_e,2*x) - 1)/(power(c_e,2*x) + 1)
end if
return result
end function

global function arccosh(object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = arccosh(x[count])
    end for
else
    if x>=1 then
        result = log(x + sqrt(x*x - 1))
    else
        result = nan
    end if
end if
return result
end function

global function arcsinh(object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = arcsinh(x[count])
    end for
else
    result = log(x + sqrt(x*x + 1))
end if
return result
end function

global function arctanh(object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = arctanh(x[count])
    end for
else
    if (x>-1 and x<1) then
        result = log(sqrt((x + 1)/(1 - x)))
    elsif (x=-1 or x=1) then
        result = inf*x
    else
        result = nan
    end if
end if
return result
end function

global function int(object x)
return floor(x)
end function

global function ln(object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = ln(x[count])
    end for
else
    if x=0 then
        result = -inf
    elsif x>0 then
        result = log(x)
    else
        result = nan
    end if
end if
return result
end function

global function logbase(atom y, object x)
object result
if sequence(x) then
    result = repeat(0,length(x))
    for count=1 to length(x) do
        result[count] = logbase(y,x[count])
    end for
else
    if (x>0 and y>0) then
        result = log(x)/log(y)
    else
        result = nan
    end if
end if
return result
end function

global function log10(object x)
return logbase(10,x)
end function

global function sortA(object x)
integer gap, j, first, last
object tempi, tempj
if sequence(x) then
    last = length(x)
    gap = floor(last / 3) + 1
    while 1 do
	first = gap + 1
	for i = first to last do
	    tempi = x[i]
	    j = i - gap
	    while 1 do
		tempj = x[j]
		if compare(tempi, tempj) >= 0 then
		    j = j + gap
		    exit
		end if
		x[j+gap] = tempj
		if j <= gap then
		    exit
		end if
		j = j - gap
	    end while
	    x[j] = tempi
	end for
	if gap = 1 then
	    return x
	else
	    gap = floor(gap / 3) + 1
	end if
    end while
else
    return x
end if
end function

global function sortD(object x)
object y
if sequence(x) then
    y = sortA(x)
    for z = 0 to (length(x)-1) do
        x[z+1] = y[length(x)-z]
    end for
end if
return x
end function

global function max(object x)
atom result
if sequence(x) then
    result = x[1]
    for count=2 to length(x) do
        if atom(x[count]) then
        if x[count]>result then
            result = x[count]
        end if
        end if
    end for
else
    result = x
end if
return result
end function

global function min(object x)
atom result
if sequence(x) then
    result = x[1]
    for count=2 to length(x) do
        if atom(x[count]) then
        if x[count]<result then
            result = x[count]
        end if
        end if
    end for
else
    result = x
end if
return result
end function

global function mean(object x)
atom result
result = 0
if sequence(x) then
    for count=1 to length(x) do
        x[count] = mean(x[count])
        result = result + x[count]
    end for
    result = result/length(x)
else
    result = x
end if
return result
end function

global function oned(object x)
sequence result
if sequence(x) then
    result = {}
    for count=1 to length(x) do
        result = result & oned(x[count])
    end for
else
    result = {x}
end if
return result
end function

global function median(object x)
atom result, average, diff, lowest
if sequence(x) then
    x = sortA(x)
    if integer(length(x)/2) then
        result = (x[length(x)/2] + x[length(x)/2+1])/2
    else
        result = x[length(x)/2+.5]
    end if
else
    result = x
end if
return result
end function

global function variance(object x)
atom result, average, average2
average2 = mean(power(x,2))
average = mean(x)
result = average2 - average * average
return result
end function

global function std_dev(object x)
atom result, var
var = variance(x)
result = sqrt(var)
return result
end function

global function fib(object x)
object result
if sequence(x) then
    result=repeat(0,length(x))
    for count=1 to length(x) do
        result[count]=fib(x[count])
    end for
elsif (integer(x) and x>=0) then
    result=int((power(c_gr,x)-power(c_grneg,x))/sqrt(5))
else
    result=nan
end if
return result
end function

global function factorial(object x)
object result
if sequence(x) then
    result=repeat(0,length(x))
    for count=1 to length(x) do
        result[count]=factorial(x[count])
    end for
elsif integer(x) then
    if x>0 then
        result=x*factorial(x-1)
    elsif x=0 then
        result=1
    else
        result=nan
    end if
else
    result=nan
end if
return result
end function

global function nPr(object a, object b)
atom result
result = factorial(a)/factorial(a-b)
return result
end function

global function nCr(object a,object b)
atom result
result = factorial(a)/(factorial(a-b)*factorial(b))
return result
end function
